#!/usr/bin/env python
# -*- coding: utf_8 -*-

# test with: static/Media/Videos/Cele/Cele_patinando_cerca_de_Venecia.mp4

import os, mmap, io
from mimetypes import guess_type
import argparse #, copy
from . import __version__ as oxi_version
       
class Smmap:
    def __init__(self, fileno, offset:int = 0, limit:int =  0):
        self._mmap = mmap.mmap(fileno, 0, flags=mmap.MAP_PRIVATE)
        self._offset = offset
        self._limit = limit
        self._mmap.seek(self._offset)

    def __getitem__(self, index):
        shifted_index = None
        if isinstance(index, slice):
            indexstart = 0 if not index.start else index.start
            indexstop = len(self._mmap) if not index.stop else index.stop
            indexstart = 0 if indexstart < 0 else indexstart
            indexstart = self._limit if indexstart > self._limit else indexstart
            indexstop = 0 if indexstop < 0 else indexstop
            indexstop = self._limit if indexstop > self._limit else indexstop
            shifted_index = slice(indexstart + self._offset, indexstop + self._offset, index.step)
        else:
            shifted_index = index
            shifted_index = 0 if shifted_index < 0 else shifted_index
            shifted_index = self._limit if shifted_index > self._limit else shifted_index
            shifted_index += self._offset
        return self._mmap.__getitem__(shifted_index)
    
    def __setitem__(self, index, value):
        shifted_index = None
        if isinstance(index, slice):
            indexstart = 0 if not index.start else index.start
            indexstop = len(self._mmap) if not index.stop else index.stop
            indexstart = 0 if indexstart < 0 else indexstart
            indexstart = self._limit if indexstart > self._limit else indexstart
            indexstop = 0 if indexstop < 0 else indexstop
            indexstop = self._limit if indexstop > self._limit else indexstop
            diff = indexstop - indexstart
            if not isinstance(value, bytes) or len(value) > diff:
                raise ValueError(f"value must be an instance of bytes with len <= to {diff}")
            for i in range(diff):
                self._mmap[indexstart + i + self._offset] = value[i]
        else:
            shifted_index = index
            shifted_index = 0 if shifted_index < 0 else shifted_index
            shifted_index = self._limit if shifted_index > self._limit else shifted_index
            shifted_index += self._offset
            self._mmap[shifted_index], = value
    
    @property
    def size(self):
        return self._limit
    
    @property
    def offset(self):
        return self._offset
    
    @property
    def mmap(self):
        return self._mmap
    
    def seek(self, num: int):
        num = self._limit if num >= self._limit else num
        self._mmap.seek(num + self._offset)

    def tell(self):
        return self._mmap.tell() - self._offset

    def read(self, nbytes: int = -1):
        if nbytes == -1 or nbytes >= (self._limit - self.tell()):
            nbytes = self._limit - self.tell()
        return self._mmap.read(nbytes)
    
    def write(self, what: bytes = b''):
        self._mmap.write(what)

    def find(self, what: bytes):
        found = self._mmap.find(what)
        return found if found == -1 else found - self._offset

    def rfind(self, what: bytes):
        found = self._mmap.rfind(what)
        return found if found == -1 else found - self._offset

    def close(self):
        self._mmap.close()

    def __del__(self):
        try:
            if not self._mmap.closed:
                self._mmap.close()
        except:
            pass


class Mp4:
    """Stub for Atom"""

class Atom:
    def __init__(self, offset: int, size: int, name: str, level: int = 0, ordinal: int = -1, container=None):
        self.offset = offset
        self.size = size
        self.name = name
        self.ordinal = ordinal
        self.level = level
        # self.contents = contents
        self.container = container
        self.children: list = []
        self._mm = None

    @property
    def contents(self):
        if self._mm:
            return self._mm
        if not self.container:
            return None
        fp = self.container.fp
        if not fp:
            return None
        self._mm = Smmap(fp.fileno(), self.offset, self.size)
        return self._mm

    def setcontents(self, new_mm):
        self._mm = new_mm

    def __repr__(self):
        # Atom ftyp @ 0 of size: 32, ends @ 32
        args = ('\t' * self.level, self.name, f"{self.offset:,}", f"{self.size:,}", f"{(self.offset + self.size):,}")
        return "{0}Atom {1} @ {2} of size: {3} , ends @ {4}".format(*args)


    def __str__(self):
        mainline = repr(self)
        lenchildren = len(self.children) if self.children else 0
        childrenlines = ''.join(map(str, self.children)) if lenchildren else ''
        return "\n".join([mainline, childrenlines])

    def load_contents(self, fp = None):
        # Retained for backward compatibility
        pass

    def load_children(self):
            container: Mp4 = self.container
            ordinal = 0
            offset = self.offset + 8
            fp = mmap.mmap(container.fp.fileno(), container.filesize, flags=mmap.MAP_PRIVATE)
            while offset < (self.size + self.offset):
                child = container._get_atom(fp, offset, ordinal, self.level + 1)
                if child.size == 0:
                    raise ValueError("Atom of size 0 not admitted out of 0 level.")
                if child.name == '\x00\x00\x00\x00':
                    offset += child.size
                    continue
                if child.name in ["trak", "mdia", "minf", "stbl"]:
                    # Known ancestor atom of stco or co64, search within it!
                    # for res in _find_atoms_ex(atom, datastream):
                    #     yield res
                    child.load_children()
                elif child.name in ["stco", "co64"]:
                    self.container.relocation_targets.append(child)
                self.children.append(child)
                ordinal += 1
                offset += child.size
          
    @property
    def boundaries(self):
        return [self.offset, self.offset + self.size]
    
    @property
    def named_boundaries(self):
        return {self.name: self.boundaries}


class Mp4:
    def __init__(self, filename):
        if not os.path.exists(filename):
            raise NameError(f"File {filename} doesn't exist.")
        if guess_type(filename)[0] not in ['video/mp4', 'video/quicktime']:
            raise TypeError(f"File {filename} is not an MP4 file.")
        self.filename = filename
        self.filesize = os.path.getsize(filename)
        self.loop = None
        self.atoms: list[Atom] = []
        self.relocation_targets: list[Atom] = []
        self.ftyp: Atom = None
        self.moov: Atom = None
        self.free: Atom = None
        self.mdat: Atom = None
        self._patched_moov: bytearray = None
        self.fp = open(self.filename, 'rb')
        self.atoms = self._collectatoms()

    def __del__(self):
        try:
            if not self.fp.closed:
                self.fp.close()
        except:
            pass

    def __str__(self):
        atomstrs = "\n".join(list(map(lambda a: "   " + str(a), self.atoms)))
        return f"""

{'_' * 80}

File: {self.filename}
Size: {self.filesize:,}
{'_' * 80}

{atomstrs}

{'_' * 80}
"""

    def __repr__(self):
        return "\n".join(self.quicklist())
    
    def _get_atom(self, fp: mmap.mmap, offset: int, ordinal: int=0, level:int=0):
        size, name = [int.from_bytes(fp[offset:offset+4], byteorder='big'), fp[offset+4:offset+8].decode('latin-1')]
        if size == 1:
            size = int.from_bytes(fp[offset+8:offset+16], byteorder='big')
        atom = Atom(offset=offset, size=size, name=name, level=level, ordinal = ordinal, container=self)
        return atom
    
    def _parse_mp4(self):
        filesize = self.filesize
        with mmap.mmap(self.fp.fileno(), filesize, flags=mmap.MAP_PRIVATE) as fp:
            ordinal = 0
            offset = 0
            while offset < filesize:
                atom = self._get_atom(fp, offset, ordinal, 0)
                if atom.size == 0:
                    atom.size = self.filesize - offset
                    yield atom
                    return
                ordinal += 1
                offset += atom.size
                if atom.name != '\x00\x00\x00\x00':
                    yield atom
                else:
                    continue

    def _collectatoms(self):
        atoms = [self._classify_atom(atom) for atom in self._parse_mp4()]
        # if self.moov:
        #     self.moov.load_children()
        return atoms
    
    def _is_relocatable(self):
        return (not not self.moov) and (not self._is_compressed()) and (self.moov.ordinal > self.mdat.ordinal)
    
    def _patch_moov(self):
        if self._patched_moov is not None or not self.moov or self._is_compressed():
            return
        offset_shift = self.moov.size
        self.moov.contents.seek(0)
        self._patched_moov = mmap.mmap(-1, self.moov.size)
        self._patched_moov.write(self.moov.contents.read())
        for atom in self.relocation_targets:
            os.sys.stderr.write(f'{atom.name} located at ' + str(atom.offset) + '\n')
            os.sys.stderr.flush()
            offset_pos_begin = atom.offset - self.moov.offset
            offset_pos = offset_pos_begin
            entry_count = int.from_bytes(self._patched_moov[offset_pos+12:offset_pos+16], byteorder='big')
            print(f"Patching atom '{atom.name}' with {entry_count} entries.")
            displacement = 4 if atom.name == 'stco' else 8 # in case atom.name == 'co64'
            for i in range(entry_count):
                offset_pos = offset_pos_begin + 16 + (i * displacement)
                current_offset = int.from_bytes(self._patched_moov[offset_pos:offset_pos+4], byteorder='big')
                new_offset = current_offset + offset_shift
                # print(f"{i}: Transferring offset {current_offset} to {new_offset}")
                self._patched_moov[offset_pos:offset_pos+4] = new_offset.to_bytes(displacement, byteorder='big')
    
    def _classify_atom(self, atom: Atom):
        # atom.container = self
        # atom.load_contents(self.fp)
        if atom.name == 'ftyp':
            self.ftyp = atom
        elif atom.name == 'moov':
            atom.load_children()
            self.moov = atom
        elif atom.name == 'free':
            self.free = atom
        elif atom.name == 'mdat':
            self.mdat = atom
        elif not hasattr(self, atom.name):
            setattr(self, atom.name, atom)
        return atom
    
    def _get_chunks(self, limit: int, stream = None, chunksize:int = None):
        if stream is None:
            stream = self.fp
        chunk_size = chunksize or mmap.PAGESIZE
        while chunk_size < (2 ** 15):
            chunk_size *= 2
        remaining = limit
        while remaining:
            chunk = stream.read(min(remaining, chunk_size))
            if not chunk:
                return
            remaining -= len(chunk)
            yield chunk

    @property
    def boundaries(self):
        return list(map(lambda a: a.named_boundaries, self.atoms))

    @property 
    def faststart(self):
        if not self._is_relocatable():
            return self.atoms
        temp_ordinal = self.mdat.ordinal
        self.mdat.ordinal = self.moov.ordinal
        self.moov.ordinal = temp_ordinal
        self.moov.setcontents(self.patched_moov)
        self.atoms.sort(key=lambda atom: atom.ordinal)
        for index, atm in enumerate(self.atoms):
            atm.contents.seek(0)
            if index == 0:
                continue
            atm.offset = self.atoms[index-1].offset + self.atoms[index-1].size
        return self.atoms

    @property
    def faststart_boundaries(self):
        return list(map(lambda a: a.named_boundaries, self.faststart))

    @property
    def patched_moov(self):
        if not self._patched_moov:
            self._patch_moov()
        return self._patched_moov
    
    def _get_moov_atom(self):
        """
        Deprecated. 'moov' is now an Mp4 attribute.
        """
        for atom in self.atoms:
            if atom.name == 'moov':
                return atom
        return None
    
    def _is_compressed(self):
        for atom in self.moov.children:
            if atom.name == 'cmov':
                print(f"Warning!\nFile {self.filename} cannot be optimized for streaming because it's compressed.")
                return True
        return False
    
    def stream(self, begin: int=0, end: int=None):
        for atm in self.faststart:
            atm.contents.seek(0)
            if atm.name not in ['mdat', 'moov']:
                yield atm.contents.read()
            else:
                chunkgen = self._get_chunks(limit=atm.size, stream=atm.contents)
                for chunk in chunkgen:
                    try:
                        yield chunk
                    except:
                        chunkgen.close()
                        break
            print(f"Mp4 '{atm.name}' stream sent succesfully ({atm.size:,} bytes).")

    def stream_range(self, begin: int=0, end: int=0):
        if not end:
            end = self.filesize - 1
        beginner, ender = {} , {}
        for atom in self.faststart:
            if begin in range(*atom.boundaries):
                beginner['name'] = atom.name
                beginner['ordinal'] = atom.ordinal
                beginner['offset'] = begin - atom.offset
            if end in range(*atom.boundaries):
                ender['name'] = atom.name
                ender['ordinal'] = atom.ordinal
                ender['offset'] = end - atom.offset
        if not ender:
            atom = self.faststart[-1]
            ender['name'] = atom.name
            ender['ordinal'] = atom.ordinal
            ender['offset'] = atom.size - 1
        if not beginner:
            atom = self.faststart[0]
            ender['name'] = atom.name
            ender['ordinal'] = atom.ordinal
            ender['offset'] = 0
        fragment = None
        for atom in self.faststart:
            if atom.ordinal < beginner.get('ordinal'):
                continue
            if atom.ordinal > ender.get('ordinal'):
                return
            if atom.ordinal == beginner.get('ordinal'):
                frag_begin = beginner.get('offset')
                if atom.ordinal == ender.get('ordinal'):
                    length = ender.get('offset') - frag_begin
                    # fragment = atom.contents[beginner.get('offset'):ender.get('offset')]
                else:
                    length = atom.size - frag_begin
                    # fragment = atom.contents[beginner.get('offset'):]
            elif atom.ordinal == ender.get('ordinal'):
                frag_begin = 0
                length = ender.get('offset')
                # fragment = atom.contents[:ender.get('offset')]
            else:
                frag_begin = 0
                length = atom.size
                # fragment = atom.contents[:]
            atom.contents.seek(frag_begin)
            gen = self._get_chunks(length, atom.contents)
            for chunk in gen:
                yield chunk

    def save(self, outputfile: str = 'out.mp4'):
        with open(outputfile, 'wb') as fp:
            for chunk in self.stream():
                fp.write(chunk)

        print(f"Mp4 file {self.filename} saved succesfully to {outputfile}.")

    def quicklist(self):
        fileline = f"File: {self.filename}"
        sizeline= f"Size: {self.filesize:,}"
        underline = "_" * max(len(fileline), len(sizeline))
        return [underline, fileline, sizeline, underline, *list(map(repr, self.atoms))]
    
def main():
    parser = argparse.ArgumentParser(description='Command line arguments for Mp4Parser', prog='mp4parser')
    parser.add_argument('-V', '--version', action='version', version=f"{parser.prog} v {oxi_version}", help=f"Shows {parser.prog} version and exits.")
    parser.add_argument('inputfile', type=str, help="Name of the file to be listed/processed.")
    parser.add_argument('-o', '--outputfile', type=str, default='out.mp4', help="Name of the file you wish to save to. Defaults to 'out.mp4'")
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-s', '--save', action="store_true", help="Use this option to save the input file to a new one.")
    group.add_argument('-l', '--list', action="store_true", help="Shows a list of the main level atoms contained in the input file.")
    group.add_argument('-L', '--full-list', action="store_true", help="Shows a detailed list of the atoms contained in the input file.")
    
    args = parser.parse_args()
    filename = args.inputfile
    
    try:
        mp4 = Mp4(filename)
    except Exception as exc:
        print(f"ERROR: {exc}")
        raise SystemExit

    if  args.save:
        mp4.save(args.outputfile)
    elif args.list:
        print('\n'.join(mp4.quicklist()))
    elif args.full_list:
        print(mp4)
    else:
        print('\n'.join(mp4.quicklist()))

if __name__ == '__main__':
    main()




#-*- coding: utf-8 -*-
import re, asyncio, platform
from functools import wraps

def is_windows():
    """
    Check if the current operating system is Windows.
    """
    return platform.system() == 'Windows'

def is_linux():
    """
    Check if the current operating system is Linux.
    """
    return platform.system() == 'Linux'

def is_mac():
    """
    Check if the current operating system is macOS.
    """
    return platform.system() == 'Darwin'
def is_unix():
    """
    Check if the current operating system is Unix-like.
    """
    return platform.system() in ['Linux', 'Darwin']

def is_posix():
    """
    Check if the current operating system is POSIX-compliant.
    """
    return os.name == 'posix'

def is_nt():
    """
    Check if the current operating system is Windows NT.
    """
    return os.name == 'nt'


######################################################################################

def to_bytes(text, encoding='utf-8'): 
    if type(text) == bytes:
        return text
    elif type(text) == int or type(text) == float:
        return str(text).encode(encoding)
    elif isinstance(text, (tuple, list)):
        return list(map(to_bytes, text))
    else:
        return str(text).encode(encoding)

######################################################################################

def safe_filename(name):
    name = name.replace(' ', '_')
    return re.sub(r'[^a-zA-Z0-9._-]', '', name)

######################################################################################

def dual_mode(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            loop = asyncio.get_running_loop()
            if loop.is_running():
                async def async_wrapper():
                    return await asyncio.to_thread(func, *args, **kwargs)
                return async_wrapper()
        except RuntimeError:
            pass
        return func(*args, **kwargs)
    
    return wrapper

######################################################################################

async def aopen(filename:str, mode:str='r'):
    coro = asyncio.to_thread(open, filename, mode)
    return await coro

######################################################################################

class SmartDict(dict):
    """SmartDict class with attributes equating dict keys"""

    def __init__(self, d: dict = None, **kw):
        super().__init__(d or {}, **kw)

    def __getattr__(self, attr):
        if attr in self:
            return self[attr]
        raise AttributeError(f"'SmartDict' object has no attribute '{attr}'")

    def __setattr__(self, attr, value):
        if hasattr(dict, attr):
            raise AttributeError(f"Cannot set attribute '{attr}': reserved name")
        self[attr] = value

    def __delattr__(self, attr):
        if attr in self:
            del self[attr]
        else:
            raise AttributeError(f"'SmartDict' object has no attribute '{attr}'")

    def copy(self):
        return SmartDict(super().copy())

######################################################################################




# -*- coding: utf-8 -*-

# This script is used to manually activate a virtual environment by modifying
# the system path and environment variables to point to the virtual environment.

import os, sys
from pathlib import Path

oxi_env = False

try:
    print("Activating virtualenv...")

    # Path to the virtual environment
    oxi_path = Path(os.path.abspath(__file__)).parent.parent.parent

    # Update sys.path to include oxi path
    sys.path.insert(0, str(oxi_path))

    oxi_env = True
    print(f"Virtual environment activated for Oxi: {str(oxi_path)}.")
except Exception as e:
    print("Virtual environment activation failed. Error: {0}".format(str(e)))




#-*- coding: utf_8  -*-

import typing, json, asyncio
from time import time, strftime
from uuid import uuid4
from . import __version__ as oxi_version

oxi_version

CONNECTING = 0
CONNECTED = 1
DISCONNECTED = 2
RESPONSE = 3

class WSDisconnection(Exception):
    def __init__(self, code: int = 1000, reason= None) -> None:
        self.code = code
        self.reason = reason or ""

class WebSocket:
    """Server side counterpart of the client WebSocket object"""

    def __init__(self, scope: dict, receive: typing.Callable, send: typing.Callable, *, heartbeat_rate = 6):
        assert scope.get("type") == "websocket"
        self._scope = scope
        self._receive = receive
        self._send = send
        self._heartbeat_rate  = heartbeat_rate
        self._pings_sent = 0
        self._ping_time = 0
        self._pong_time = 0
        self._pings = {}
        self._pongs = {}
        self._roundtrips = {}
        self.client_state = CONNECTING
        self.application_state = CONNECTING

    async def start_heartbeat(self):
        async def do_heartbeating(interval: int):
            while True:
                try:
                    await asyncio.sleep(interval)
                    await self.send_ping()
                    if self.application_state != CONNECTED:
                        break
                    if self.client_state != CONNECTED:
                        break
                except:
                    await self.close()
                    break

        heartbeat_interval = 60 // self._heartbeat_rate
        await do_heartbeating(heartbeat_interval)

    async def receive(self) -> dict:
        """
        Receive ASGI websocket messages.
        """
        if self.client_state == CONNECTING:
            message = await self._receive()
            message_type = message.get("type")
            if message_type != "websocket.connect":
                raise RuntimeError(
                    'Expected ASGI message "websocket.connect", '
                    f"but got {message_type!r}"
                )
            self.client_state = CONNECTED
            return message
        elif self.client_state == CONNECTED:
            message = await self._receive()
            message_type = message.get("type", "websocket.receive")
            if message_type not in {"websocket.receive", "websocket.disconnect", "websocket.ping", "websocket.pong"}:
                raise RuntimeError(
                    'Expected ASGI message "websocket.receive" or '
                    f'"websocket.disconnect" or "websocket.ping" or "websocket.pong", but got {message_type!r}'
                )
            if message_type == "websocket.disconnect":
                self.client_state = DISCONNECTED
            elif message_type == "websocket.ping":
                await self.send_pong(message)
            elif message_type =="websocket.pong":
                payload = message.get('bytes').decode('utf_8')
                if payload in self._pings.keys():
                    start_time = self._pings.get(payload)
                    end_time = time()
                    self._pongs[payload] = end_time
                    roundtrip_time = end_time - start_time
                    self._roundtrips[payload] = roundtrip_time
                    print(f"Roundtrip time for ping-pong '{payload}' is: {roundtrip_time} secs.") 
            return message
        else:
            raise RuntimeError(
                'Cannot call "receive" once a disconnect message has been received.'
            )

    async def send(self, message: dict) -> None:
        """
        Send ASGI websocket messages, ensuring valid state transitions.
        """
        if self.application_state == CONNECTING:
            message_type = message.get("type")
            if message_type not in {
                "websocket.accept",
                "websocket.close",
                "websocket.http.response.start",
            }:
                raise RuntimeError(
                    'Expected ASGI message "websocket.accept",'
                    '"websocket.close" or "websocket.http.response.start",'
                    f"but got {message_type!r}"
                )
            if message_type == "websocket.close":
                self.application_state = DISCONNECTED
            elif message_type == "websocket.http.response.start":
                self.application_state = RESPONSE
            else:
                self.application_state = CONNECTED
                # await self.start_heartbeat()
            await self._send(message)
        elif self.application_state == CONNECTED:
            message_type = message.get("type")
            if message_type not in {"websocket.send", "websocket.close", "websocket.ping", "websocket.pong"}:
                raise RuntimeError(
                    'Expected ASGI message "websocket.send" or "websocket.close" or "websocket.ping" or "websocket.pong"'
                    f"but got {message_type!r}"
                )
            if message_type == "websocket.close":
                self.application_state = DISCONNECTED
            try:
                await self._send(message)
            except OSError:
                self.application_state = DISCONNECTED
                raise WSDisconnection(code=1006)
            except Exception as exc:
                self.application_state = DISCONNECTED
                print(f"Exception inf {__file__}, line 115: {repr(exc)}")
        elif self.application_state == RESPONSE:
            message_type = message["type"]
            if message_type != "websocket.http.response.body":
                raise RuntimeError(
                    'Expected ASGI message "websocket.http.response.body", '
                    f"but got {message_type!r}"
                )
            if not message.get("more_body", False):
                self.application_state = DISCONNECTED
            await self._send(message)
        else:
            print(f"""{RuntimeError('Cannot call "send" once a close message has been sent.')}""")
            raise RuntimeError('Cannot call "send" once a close message has been sent.')

    async def accept(
        self,
        subprotocol= None,
        # headers: typing.Iterable[tuple[bytes, bytes]] | None = None,
        headers = None,
    ) -> None:
        headers = headers or []

        if self.client_state == CONNECTING:
            # If we haven't yet seen the 'connect' message, then wait for it first.
            recvd = await self.receive()
            print(f"Received: {repr(recvd)}")
        await self.send(
            {"type": "websocket.accept", "subprotocol": subprotocol, "headers": headers}
        )
        try:
            asyncio.create_task(self.start_heartbeat())
        except Exception as exc:
            print(f"Exception {exc} happened in heartbeat coroutine.")

    def raise_disconnection(self, message: dict) -> None:
        message_type = message.get("type")
        if not message_type:
            return
        if message_type == "websocket.disconnect":
            # self.application_state = DISCONNECTED
            raise WSDisconnection(message.get("code", 1000), message.get("reason", ''))

    async def receive_text(self) -> str:
        if self.application_state != CONNECTED:
            raise RuntimeError(
                'WebSocket is not connected. Need to call "accept" first.'
            )
        message = await self.receive()
        msg_type = message.get("type")
        if msg_type== "websocket.ping" or msg_type == "websocket.pong":
            return ''
        self.raise_disconnection(message)
        # if self.application_state == DISCONNECTED:
        #     return ''
        return typing.cast(str, message.get("text"))

    async def receive_bytes(self) -> bytes:
        if self.application_state != CONNECTED:
            raise RuntimeError(
                'WebSocket is not connected. Need to call "accept" first.'
            )
        message = await self.receive()
        msg_type = message.get("type")
        if msg_type== "websocket.ping" or msg_type == "websocket.pong":
            return b''
        self.raise_disconnection(message)
        return typing.cast(bytes, message.get("bytes"))

    async def receive_json(self, mode: str = "text") -> typing.Any:
        if mode not in {"text", "binary"}:
            raise RuntimeError('The "mode" argument should be "text" or "binary".')
        if self.application_state != CONNECTED:
            raise RuntimeError(
                'WebSocket is not connected. Need to call "accept" first.'
            )
        message = await self.receive()
        msg_type = message.get("type")
        if msg_type== "websocket.ping" or msg_type == "websocket.pong":
            return {}

        self.raise_disconnection(message)
        if not hasattr(message, 'get'):
            return
        keys = message.keys()
        if 'bytes' in keys:
            text = message.get('bytes', b'').decode('utf_8')
        elif 'text' in keys:
            text = message.get('text', '')
        else:
            return
        
        try:
            jsontxt = json.loads(text)
        except:
            jsontxt = ''
        return jsontxt

    async def send_text(self, data: str) -> None:
        await self.send({"type": "websocket.send", "text": data})

    async def send_bytes(self, data: bytes) -> None:
        await self.send({"type": "websocket.send", "bytes": data})

    async def send_json(self, data: typing.Any, mode: str = "text") -> None:
        if mode not in {"text", "binary"}:
            raise RuntimeError('The "mode" argument should be "text" or "binary".')
        text = json.dumps(data, separators=(",", ":"), ensure_ascii=False)
        if mode == "text":
            await self.send({"type": "websocket.send", "text": text})
        else:
            await self.send({"type": "websocket.send", "bytes": text.encode("utf_8")})

    async def send_ping(self):
        # await asyncio.sleep(0.1)
        if not self._pings_sent:
            print("\nStarting hearbeat.")
            print()
        self._pings_sent += 1
        self._ping_time = time()
        payload = uuid4().hex.encode('utf_8')
        # payload = b''
        print(f"Sending ping number {self._pings_sent} at {strftime('%X')} with payload: '{payload.decode()}'")
        if len(self._pings) > 10:
            self._pings = {}
            self._pongs = {}
            self._roundtrips = {}
        self._pings[payload.decode('utf_8')] = time()    
        try:
            await self.send({"type": "websocket.ping", "bytes": payload})
        except:
            await self.close()
    
    async def send_pong(self, message):
        if hasattr(message, 'get'):
            payload = message.get('bytes', b'')
        else:
            payload = b''
        await self.send({"type": "websocket.pong", "bytes": payload})

    # async def close(self, code: int = 1000, reason: str | None = None) -> None:
    async def close(self, code: int = 1000, reason= None) -> None:
        await self.send(
            {"type": "websocket.close", "code": code, "reason": reason or ""}
        )
        self.application_state = DISCONNECTED

    @property
    def closed(self):
        return self.application_state == DISCONNECTED or self.client_state == DISCONNECTED


# -*- coding: utf-8 -*-

import asyncio, os, random, subprocess, re, mimetypes
from typing import Callable

try:
    from . import  __version__ as oxi_version, __oxi_port__ as oxi_port, __oxi_host__ as oxi_host
    from .utils import dual_mode, is_windows, is_linux, is_unix
except ImportError:
    from activate_this import oxi_env
    if not oxi_env:
        raise ImportError("Oxi environment not activated. Please activate the virtual environment.")
    from oxi import  __version__ as oxi_version, __oxi_port__ as oxi_port, __oxi_host__ as oxi_host
    from oxi.utils import dual_mode, is_linux, is_windows, is_unix

async def is_static(resource:str) -> bool:
    """
    Check if the resource exists in the file system.
    """
    return await asyncio.to_thread(os.path.exists, resource)

async def is_file(resource:str) -> bool:
    """
    Check if the resource is a regular file.
    """
    exists = await asyncio.to_thread(os.path.exists, resource)
    return exists and await asyncio.to_thread(os.path.isfile, resource)

async def is_dir(resource:str) -> bool:
    """
    Check if the resource is a directory.
    """
    exists = await asyncio.to_thread(os.path.exists, resource)
    return exists and await asyncio.to_thread(os.path.isdir, resource)

async def is_exe(resource:str) -> bool:
    """
    Check if the resource is an executable file.
    """
    exists = await asyncio.to_thread(os.path.exists, resource)
    return exists and await asyncio.to_thread(os.access, resource, os.X_OK)

async def is_cgi_exe(resource:str, cgi_dir:str="cig-bin") -> bool:
    """
    Check if the resource is a CGI executable.
    """
    exists = await asyncio.to_thread(os.path.exists, resource)
    if not exists:
        return False
    is_exe = await is_exe(resource)
    if not is_exe:
        return False
    if not cgi_dir in resource:
        return False
    return True

enctypes: tuple = (b"application/x-www-form-urlencoded", 
                   b"multipart/form-data", 
                   b"application/json",
                   b"text/plain")

strftime_template = "%a, %d  %b %Y %X %Z"

original_zen = """Gur Mra bs Clguba, ol Gvz Crgref

Ornhgvshy vf orggre guna htyl.
Rkcyvpvg vf orggre guna vzcyvpvg.
Fvzcyr vf orggre guna pbzcyrk.
Pbzcyrk vf orggre guna pbzcyvpngrq.
Syng vf orggre guna arfgrq.
Fcnefr vf orggre guna qrafr.
Ernqnovyvgl pbhagf.
Fcrpvny pnfrf nera'g fcrpvny rabhtu gb oernx gur ehyrf.
Nygubhtu cenpgvpnyvgl orngf chevgl.
Reebef fubhyq arire cnff fvyragyl.
Hayrff rkcyvpvgyl fvyraprq.
Va gur snpr bs nzovthvgl, ershfr gur grzcgngvba gb thrff.
Gurer fubhyq or bar-- naq cersrenoyl bayl bar --boivbhf jnl gb qb vg.
Nygubhtu gung jnl znl abg or boivbhf ng svefg hayrff lbh'er Qhgpu.
Abj vf orggre guna arire.
Nygubhtu arire vf bsgra orggre guna *evtug* abj.
Vs gur vzcyrzragngvba vf uneq gb rkcynva, vg'f n onq vqrn.
Vs gur vzcyrzragngvba vf rnfl gb rkcynva, vg znl or n tbbq vqrn.
Anzrfcnprf ner bar ubaxvat terng vqrn -- yrg'f qb zber bs gubfr!"""

zen = """
The Zen of Python, by Tim Peters

Beautiful is better than ugly.
Explicit is better than implicit.
Simple is better than complex.
Complex is better than complicated.
Flat is better than nested.
Sparse is better than dense.
Readability counts.
Special cases aren't special enough to break the rules.
Although practicality beats purity.
Errors should never pass silently.
Unless explicitly silenced.
In the face of ambiguity, refuse the temptation to guess.
There should be one-- and preferably only one --obvious way to do it.
Although that way may not be obvious at first unless you're Dutch.
Now is better than never.
Although never is often better than *right* now.
If the implementation is hard to explain, it's a bad idea.
If the implementation is easy to explain, it may be a good idea.
Namespaces are one honking great idea -- let's do more of those!
"""


class ProtocolFactory: 
    """
    ProtocolFactory class to handle the server protocol.
    """
    
    success_line = "HTTP/1.1 200 OK\r\n"
    error_line = "HTTP/1.1 500 Internal Server Error\r\n"
    not_found_line = "HTTP/1.1 404 Not Found\r\n"

    @property
    def base_dir(self) -> str:
        if len(self._base_dir) == 0:
            return ''
        return f"{os.path.sep}{self._base_dir}"

    @base_dir.setter
    def base_dir(self, value: str) -> None:
        self._base_dir = value

    @property
    def full_base_dir(self) -> str:
        return f"{os.getcwd()}{self.base_dir}"

    def __init__(self, app: Callable = None, *, base_dir: str = 'static'):
        self.app = app
        if self.app:
            self.app._protocol = self

        self._base_dir = base_dir
        self.version = oxi_version
        self.zen = zen
        self.original_zen = original_zen
        self.strftemplate = strftime_template
        self.enctypes = enctypes
        self.cgi_dir = "cgi-bin"

    async def __call__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        """
        Handle the incoming request and send a response.
        """
        try:
            method, path, protocol = await self.get_request_line(reader)
        except ValueError as e:
            print(f"Error parsing request line: {e}")
            return await self.send_response(writer, status_code=400, reason="Bad Request", msg=str(e))
        print(f"Received request: {method} {path} {protocol}")
        if method not in ["HEAD", "GET", "POST", "PUT", "PATCH", "DELETE"]:
            print(f"Method {method} not allowed.")
            writer.write(self.not_found_line.encode("utf-8"))
            writer.write(b"Content-Type: text/plain\r\n")
            response = b"405 Method Not Allowed\r\n\r\n"
            writer.write(f"Content-Length: {len(response)}\r\n\r\n".encode("utf-8"))
            writer.write(response)
            await writer.drain()
            writer.close()
            await writer.wait_closed()
            return
 
        if method == "GET":
            if path == "/oxiserver_demo":
                print(f"Serving Oxi Server demo page.")
                return await self.oxiserver_demo(writer=writer)
            else:
                fullpath = os.path.join(self.full_base_dir, path.lstrip("/").replace("/", os.path.sep)) 
                if await is_file(fullpath):
                    print(f"Serving file: {fullpath}")
                    return await self.send_file(writer=writer, fullpath=fullpath)
                else:
                    print(f"File not found: {path}")
                    return await self.send_response(writer=writer, status_code=404, reason="Not Found", msg=f"File '{path}' not found.")  
            
    async def send_file(self, writer: asyncio.StreamWriter, fullpath: str) -> None:
        with open(fullpath, "rb") as f:
            data = f.read()
        content_length = len(data)
        mimetype = mimetypes.guess_type(fullpath)[0]
        content_type = mimetype or "application/octet-stream"
        writer.write(self.success_line.encode("utf-8"))
        writer.write(f"Content-Type: {content_type}\r\n".encode("utf-8"))
        writer.write(f"Content-Length: {content_length}\r\n".encode("utf-8"))
        writer.write(b"\r\n")
        await writer.drain()
        writer.write(data)
        await writer.drain()
        writer.close()
        await writer.wait_closed()

    async def get_request_line(self, reader: asyncio.StreamReader) -> tuple[str]:
        """
        Read the request line from the client.
        """
        request_line = await reader.readuntil(b"\r\n")
        if not request_line:
            raise ValueError("Request line is empty.")
        request_line = request_line.rstrip(b"\r\n")
        if not request_line:
            raise ValueError("Request line is empty after stripping.")
        request_line = request_line.decode("utf-8").strip()
        if not request_line:
            raise ValueError("Request line is empty after decoding.")
        try:
            method, path, protocol = re.split(r"\s+", request_line)
        except ValueError:
            raise ValueError("Request line is malformed. Unable to split into method, path, and protocol.")
        if not method or not path or not protocol:
            raise ValueError("Request line is malformed. Missing method, path, or protocol.")
        return method.upper(), path, protocol

    async def send_response(self, writer: asyncio.StreamWriter, status_code: int=404, reason: str = "Not found", msg:str="") -> None:
        status_line = f"HTTP/1.1 {status_code} {reason}\r\n"
        writer.write(status_line.encode("utf-8"))
        writer.write(b"Content-Type: text/html\r\n")
        realmsg = msg + '\r\n' if len(msg) else ''
        response = f"""
<html>
<head><title>{status_code} {reason}</title></head>
<body style="background-color: #f0f0f0; font-family: Helvetica, Arial, sans-serif;">
<center><h1>{status_code} {reason}</h1></center>
<hr><center>oxi/{oxi_version}</center>
<p>&nbsp;</p>
<p>{realmsg}</p>
</body>
</html>"""
        # response = f"{status_code} {reason}\r\n{realmsg}\r\n"
        writer.write(f"Content-Length: {len(response)}\r\n\r\n".encode("utf-8"))
        writer.write(response.encode("utf-8"))
        await writer.drain()
        writer.close()
        await writer.wait_closed()
        
    async def oxiserver_demo(self, writer: asyncio.StreamWriter = None) -> None:
        """
        Send a demo HTML page to the client.
        """
        if writer is None:
            raise ValueError("writer is None. Please provide a valid writer.")
        # Let's begin slowly
        exists = await is_static("./static/img")
        if not exists:
            msg = "Static directory not found. Please create a static directory with images."
            return await self.send_response(writer=writer, status_code=404, reason="Not Found", msg=msg)
        img_src = random.choice(os.listdir("./static/img"))
        newzen = random.choice([self.zen, self.original_zen]).replace("\n", "<br>")
        html = f"""
        <html>
            <head>
                <title>Oxi Server</title>
                <!--<link rel="stylesheet" href="static/css/style.css">-->
            </head>
            <body style="background-color: #f0f0f0; font-size: 120%; font-family: Helvetica, Arial, sans-serif;">
                <div style="text-align: center;">
                    <h1 style="color: steelblue;">Welcome to <strong style="color: red;">Oxi Server</strong>!</h1>
                    <img src="/img/{img_src}" alt="Random Image" width="20%" style="border-radius: 20px;">
                </div>
                <h2>Version: <strong>{self.version}</strong></h2>
                <hr/>
                <p>{newzen}</p>
                <script>
                    setInterval(() => location.href = location.href, 10000);
                </script>
            </body>
        </html>
"""
        content_length = len(html)
        print(f"Writing to socket {content_length} bytes.\n")
        content_type = "text/html; charset=utf-8"
        writer.write(self.success_line.encode("utf-8"))
        writer.write(f"Content-Type: {content_type}\r\n".encode("utf-8"))
        writer.write(f"Content-Length: {content_length}\r\n".encode("utf-8"))
        writer.write(b"\r\n")
        writer.write(html.encode("utf-8"))
        writer.write(b"\r\n")
        await writer.drain()
        writer.close()
        await writer.wait_closed()


async def run_dev_server(protocol: Callable = ProtocolFactory(), host:str=oxi_host, port:int=oxi_port) -> None:
    server = await asyncio.start_server(protocol, host, port, reuse_address=True, reuse_port=True)
    subprocess.run("clear")
    print(f"\n Oxi Server running at {host}:{port}\n")
    async with server:
        try:
            await server.serve_forever()
        except KeyboardInterrupt:
            print("\n Oxi Server stopped by user")
        except Exception as e:
            print(f"\n Oxi Server stopped with error: {e}")
        finally:            
            await server.wait_closed()
            print("\n Oxi Server closed")

async def main():
    protocol = ProtocolFactory(base_dir='')
    await run_dev_server(protocol, host=oxi_host, port=oxi_port)

if __name__ == "__main__":
    asyncio.run(main())



# -*- coding: utf-8 -*-
from activate_this import oxi_env

if oxi_env:
    from oxi import __version__ as oxi_version
from oxi.tests.first import first
from oxi.tests.second import second

response = first + second

print(f"OXI version: {oxi_version} sends unviersal response: {response}")




from setuptools import setup, find_packages
from oxi import (__version__, __author__, __email__, __license__, 
                 __url__, __description__, __copyright__)

setup(
    name='oxi',
    version=__version__,
    description=__description__,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author=__author__,
    author_email=__email__,
    url=__url__,
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
    entry_points={
        'console_scripts': [
            'mp4parser=oxi.mp4parser:main',
        ],
    },
    license='MIT',
    license_files=('LICENSE',),    
)